#!/usr/bin/env python
# -*- coding: utf-8 -*-

import maya.cmds as cmds
import hmScripts.hm_global_variables as nc
import hmScripts.Main.hmBaseObject as hmNode
import hmScripts.Main.hmUtility as util
from hmScripts.Main.hmNodeConnector import HmOperationConnect




class HmOperatorObject(object):
	"""
	The HmOperatorObject is a virtual object that
	is created for managing the various elements of a hm operation.
	"""

	# Attributes
	in_mesh_attr = nc.hm_in_mesh
	out_mesh_attr = nc.hm_out_mesh
	inter_curve_attr = nc.hm_inter_curve

	# Define the valid type of objects that we can handle
	OBJECT_TYPES = (nc.hm_in_mesh, nc.hm_out_mesh, nc.hm_inter_curve)

	# Internal options
	VERBOSE = 0 #From 0 to 2

	def __init__(self, input_obj, hm_type):
		'''
		Store all the objects necessary to the usage of the class.
		We can initialize the structure with one of these type of objects:
			hm_in_mesh
			hm_out_mesh
			hm_inter_curve

		Remember that the object type is defined by the attribute that we create in the
		creation process.


		Args:
		    input_obj: a string, rapresenting a maya object used in an hm network
		    hm_type: a string, rapresenting the type of hm object

		Returns:
			input_obj

		'''
		# Here we store the network objects
		self.input_meshes = []
		self.smooth_nodes = []
		self.intermediate_meshes = []
		self.intersection_node = None
		self.override_nodes = []
		self.intersection_curves = []
		self.feedback_curves = []
		self.offset_node = None
		self.compound_node = None
		self.result_mesh =  None

		# ============== FIND MAIN NODES ==============

		# Input mesh case
		if hm_type == self.OBJECT_TYPES[0]:
			self.__search_nodes_from_input_mesh(input_obj)

		# Output mesh case
		elif hm_type == self.OBJECT_TYPES[1]:
			self.__search_nodes_from_output_mesh(input_obj)

		# Curve case
		elif hm_type == self.OBJECT_TYPES[2]:
			self.__search_nodes_from_curve(input_obj)


		# ============== FINISH FINDING  ==============
		# Store all the input meshes
		self.__store_input_meshes_from_intersect()

		# Store all the smooth nodes
		self.__store_smooth_nodes_from_input_meshes()

		# Store all the feedback_curves
		self.__store_feedback_curves_from_offset()

		# Store the result mesh
		self.__store_result_mesh_from_compound()

		# Print message
		if self.VERBOSE:
			found_list = [self.input_meshes,
						  self.intersection_node.get_name(),
						  self.offset_node.get_name(),
						  self.compound_node.get_name(), self.result_mesh.get_name() ]
			print 'Network object initialized correctly.\n %s' % found_list

	def __repr__(self):
		return 'HmOperatorObject: %s' % self.get_network_name()

	def print_stored_data(self):
		'''
		Print the stored data as strings
		'''
		print '======== IN MESHES ========'
		print 'input meshes: ', len(self.input_meshes)
		for mesh in self.input_meshes:
			print 'input mesh: ', mesh.get_name()

		print '======== NODES ========'
		print 'Intersection node: ', self.intersection_node.get_name()
		print 'Offset node: ', self.offset_node.get_name()
		print 'Compound node: ',self.compound_node.get_name()
		print 'Result Mesh: ',self.result_mesh.get_name()



	#===========================================================
	#=============== 3 MAIN SEARCH METHODS =====================
	#===========================================================
	def __search_nodes_from_input_mesh(self, input_mesh):
		'''
		We search from the input mesh all the main nodes involved in the hm operation.
		Keep in mind that if the input mesh has the output to more than one network the search will fail.

		Args:
		    input_mesh: The start mesh for moving between connections

		Returns:
		    None
		'''
		if self.VERBOSE > 1: print 'calling __search_nodes_from_input_mesh'

		# Initialize the first object
		start_mesh = hmNode.HmBaseObject(input_mesh)

		# === Regular input mesh case
		if start_mesh.has_attribute(nc.smooth_attr_name):
			# Get the poly smooth
			poly_smooth = start_mesh.get_connected(nc.smooth_attr_name, 'out') [0][0]

			# Get the intermediate mesh
			interm_mesh = poly_smooth.get_connected('output', 'out') [0][0]

			# Add to the class list
			# self.intermediate_meshes.append(interm_mesh)



		# === Result mesh as input mesh case
		else:
			interm_mesh = start_mesh

		if self.VERBOSE > 1: print '%s initialized as start mesh.\nWe need to query the connection outMesh for getting the intersection Node.' % start_mesh.get_name()


		# Get the Intersector, make sure that we get only the righ type
		self.intersection_node = interm_mesh.get_connected('outMesh', 'out', node_type = 'pfxToon') [0][0]

		# Get the intersection curves
		int_data = self.intersection_node.get_connected('outMainCurves', 'out')
		self.intersection_curves = [int_data[x][0] for x in range(len(int_data)) ]

		# Get the offset node by querying the first curve
		self.offset_node = self.intersection_curves[0].get_connected('local', 'out') [0][0]

		# Get the compound node by querying the boolean operation connection
		self.compound_node = self.offset_node.get_connected(nc.boolean_operation, 'out') [0][0]


	def __search_nodes_from_output_mesh(self, input_mesh):
		'''
		Write here what the function does.

		Args:
		    input_mesh: The start mesh for moving between connections

		Returns:
		    None
		'''

		if self.VERBOSE > 1: print 'calling __search_nodes_from_output_mesh'


		# Initialize the first object
		start_mesh = hmNode.HmBaseObject(input_mesh)

		# Get che compound node
		choice = start_mesh.get_connected('inMesh', 'in') [0][0]

		# mesh_connection = start_mesh.get_connected('inMesh', 'in') [0][0]
		# if mesh_connection.get_type() == 'choice':
		# 	choice = mesh_connection
		# 	print 'choiche found!'
		# elif mesh_connection.get_type() == 'polyMirror':
		# 	choice = mesh_connection.get_connected('inputPolymesh', 'in') [0][0]
		# 	print ' poly mirror found!'
		# else:
		# 	print 'esticazzi'
		# 	found_type = mesh_connection.get_type()
		# 	cmds.error('Wrong node type from %s. Found %s, expected \"choice\"' % (start_mesh.get_name(), found_type) )

		self.compound_node = choice.get_connected('input[0]', 'in') [0][0]


		# Get the offset by querying the operation attribute
		self.offset_node = self.compound_node.get_connected(nc.compound_operation, 'in') [0][0]

		# Get the intersection curves
		int_data = self.offset_node.get_connected(nc.in_curves, 'in')
		self.intersection_curves = [ int_data[x][0] for x in range( len(int_data) ) ]

		# Get the intersection node
		self.intersection_node = self.intersection_curves[0].get_connected('create', 'in', node_type = 'pfxToon') [0][0]


	def __search_nodes_from_curve(self, input_curve):
		'''
		Write here what the function does.

		Args:
		    input_mesh: The start mesh for moving between connections

		Returns:
		    None
		'''

		if self.VERBOSE > 1: print 'calling __search_nodes_from_curve'

		# Initialize the first object
		start_curve = hmNode.HmBaseObject(input_curve)

		# Get the offset node
		self.offset_node = start_curve.get_connected('create', 'in') [0][0]

		# Get the intersection curves
		int_data = self.offset_node.get_connected(nc.in_curves, 'in')
		self.intersection_curves = [ int_data[x][0] for x in range( len(int_data) ) ]

		# Get the intersection node
		self.intersection_node = self.intersection_curves[0].get_connected('create', 'in', node_type = 'pfxToon') [0][0]


		# Get the compound node by querying the boolean operation connection
		self.compound_node = self.offset_node.get_connected(nc.boolean_operation, 'out') [0][0]


	#===========================================================
	#===================== FINISH SEARCH =======================
	#===========================================================
	def __store_input_meshes_from_intersect(self):
		'''
		Once we have the intersection node, we can easily find all the input meshes
		'''
		if self.VERBOSE > 1: print 'calling __store_input_meshes_from_intersect'

		input_result = self.intersection_node.get_connected('inputSurface', 'in')

		# Cycle on the result to get only the nodes connected on the matrix
		for connected in input_result:
			if connected[1] == 'worldMatrix':
				# Store the transforms of the meshes
				self.input_meshes.append(connected[0])


	def __store_smooth_nodes_from_input_meshes(self):
		'''
		Once we have the input meshes list it is easy to save all the smooth nodes
		'''
		if self.VERBOSE > 1: print 'calling __store_smooth_nodes_from_input_meshes'

		# Cycle on every input mesh for getting the smooth node
		for mesh in self.input_meshes:
			# If the mesh has the attribute
			if mesh.has_attribute(nc.smooth_attr_name):
				# By querying the hmDivision attribute we do not need to go to shape level
				smooth_data = mesh.get_connected(nc.smooth_attr_name, 'out')

				# If there is a connection
				if smooth_data:
					self.smooth_nodes.append(smooth_data[0][0])


	def __store_feedback_curves_from_offset(self):
		'''
		Once we have the offset node, it is easy to save all the offset curves
		'''

		if self.VERBOSE > 1: print 'calling __store_feedback_curves_from_offset'

		# Get the out curves, we are getting the shapes.
		# It' important to specify the node type since the connection is also going to compound
		curve_result = self.offset_node.get_connected(nc.out_curves, 'out', node_type = 'nurbsCurve')
		# Store in the var just the hmNode object, but at transform level
		self.feedback_curves =  [curve_result[x][0].get_transform() for x in range(len(curve_result)) ]


	def __store_result_mesh_from_compound(self):
		'''
		store the result mesh transform from the compound node
		'''
		# Get the result mesh
		choice = self.compound_node.get_connected('outputMesh', 'out')[0][0]
		shape = choice.get_connected('output', 'out')[0][0]
		self.result_mesh = shape.get_transform()


	#===========================================================
	#======================= GET METHODS =======================
	#===========================================================


	# ============== NODE HIERARCHY ==============

	def get_input_meshes(self):
		'''
		Return the input meshes involved in this operation

		Args:
			None

		Returns:
		    The input meshes as hmNode object
		'''

		# This shit is dirty... for some reason it was losing hte data after the handler loading
		self.input_meshes = []
		self.__store_input_meshes_from_intersect()


		return self.input_meshes

	def get_smooth_nodes(self):
		'''
		Return the smooth nodes involved in this operation

		Args:
			None

		Returns:
		    The smooth nodes as list[] of hmNode objects
		'''
		return self.smooth_nodes

	def get_intersection_node(self):
		'''
		Return the intersection node involved in this operation

		Args:
			None

		Returns:
		    The intersection node as hmNode object
		'''
		return self.intersection_node

	def get_intersection_curves(self):
		'''
		Return the intersection curves involved in this operation

		Args:
			None

		Returns:
		    The intersection curves as list[] of hmNode objects

		'''
		return [x.get_transform() for x in self.intersection_curves]

	def get_offset_curves(self):
		'''
		Return the offset curves involved in this operation

		Args:
			None

		Returns:
		    The offset curves as list[] of hmNode objects


		'''
		return self.feedback_curves


	def get_override_nodes(self):
		'''
		Return the intersection node involved in this operation

		Args:
			None

		Returns:
		    The override nodes as as list[] of hmNode objects


		'''
		return self.override_nodes

	def get_offset_node(self):
		'''
		Return the offset node involved in this operation

		Args:
			None

		Returns:
		    The offset node as hmNode object


		'''
		return self.offset_node

	def get_compound_node(self):
		'''
		Return the compound node involved in this operation

		Args:
			None

		Returns:
		    The compound node as hmNode object


		'''
		return self.compound_node

	def get_result_mesh(self):
		'''
		Return the compound node involved in this operation

		Args:
			None

		Returns:
		    The result mesh as hmNode object


		'''
		return self.result_mesh


	# ============== TRANSFORM HIERARCHY ==============
	def get_network_object(self):
		'''
		Write here what the function does.

		Args:
		    None

		Returns:
		    The hmNetwork as hmNode object
		'''
		return self.intersection_node.get_transform().get_transform()

	def get_network_name(self):
		'''
		Write here what the function does.

		Args:
		    None

		Returns:
		    This name of the group containing the intersection and the curves
		'''
		return self.intersection_node.get_transform().get_transform().get_name()

	def get_curves_grp_name(self):
		return self.feedback_curves[0].get_transform().get_name()

	#===========================================================
	#=================== EDITING METHODS ======================= these methods will be supported on 2.5 version
	#=========================================================== where we can add the multi mesh support, maybe not the attribute override yet

	def add_new_mesh(self):
		'''
		Write here what the function does.

		Args:
		    param1: This is the first param.
		    param2: This is a second param.

		Returns:
		    This is a description of what is returned.

		Raises:
		    KeyError: Raises an exception.
		'''
		NotImplemented


	def remove_mesh(self, mesh):
		'''
		Write here what the function does.

		Args:
		    param1: This is the first param.
		    param2: This is a second param.

		Returns:
		    This is a description of what is returned.

		Raises:
		    KeyError: Raises an exception.
		'''
		NotImplemented


	def add_attribute_override(self, intersection):
		'''
		Write here what the function does.

		Args:
		    param1: This is the first param.
		    param2: This is a second param.

		Returns:
		    This is a description of what is returned.

		Raises:
		    KeyError: Raises an exception.
		'''
		NotImplemented


	def remove_attribute_override(self, intersection):
		'''
		Write here what the function does.

		Args:
		    param1: This is the first param.
		    param2: This is a second param.

		Returns:
		    This is a description of what is returned.

		Raises:
		    KeyError: Raises an exception.
		'''
		NotImplemented


	def update_intersections(self):
		"""
		This method delete the current intersection curves, and
		re create them.

		We will remove it when the custom intersection node will be ready.
		"""

		# Delete the stored intersection curves
		#
		curves_to_del = self.get_intersection_curves() + self.get_offset_curves()

		for curve in curves_to_del:
			cmds.delete( curve.get_name() )


		# # Call the connection from the node connector
		#
		HmOperationConnect.create_and_connect_intersection_curves(	self.intersection_node.get_name(),
																	self.offset_node.get_name(),
																	self.compound_node.get_name(),
																	self.get_network_name(),
																	)


	#===========================================================
	#===================== CLEAN METHODS =======================
	#===========================================================
	def delete(self, keep_mesh):
		'''
		Delete all the hm nodes except the result of the current loaded operators network

		Args:
		    delete_input: If on delete the input meshes after the _ing has done

		Returns:
		    None
		'''

		net_obj = self.get_network_object()
		net_obj.delete()

		offset_obj = self.get_offset_node()
		offset_obj.delete()


		for mesh in self.input_meshes:
			if keep_mesh:
				util.clean_object(mesh)
			else:
				mesh.delete()




